﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [StorableClass]
  [Item("AfterCrossoverOperator", "Performs an action after the crossover operator is applied.")]
  public class BeforeManipulatorOperator<T> : EvolutionTrackingOperator<T>, IManipulatorOperator<T> where T : class,IItem {

    private const string ChildParameterName = "Child";

    public ILookupParameter<T> ChildParameter {
      get { return (ILookupParameter<T>)Parameters[ChildParameterName]; }
    }

    protected BeforeManipulatorOperator(BeforeManipulatorOperator<T> original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BeforeManipulatorOperator<T>(this, cloner);
    }

    public BeforeManipulatorOperator() {
      Parameters.Add(new LookupParameter<T>(ChildParameterName));
    }

    public override IOperation Apply() {
      var vChild = (IGenealogyGraphNode<T>)GenealogyGraph[ChildParameter.ActualValue];
      if (vChild.Rank.IsAlmost(Generations.Value + 1)) {
        // if the graph already contains a vertex representing the child, it means that the child is a product of crossover
        // when mutation follows after crossover, some changes need to be made to the graph to maintain consistency
        var vClone = new GenealogyGraphNode<T> {
          Content = (T)ChildParameter.ActualValue.Clone(),
          Rank = vChild.Parents.Last().Rank + 0.5
        };
        GenealogyGraph.AddVertex(vClone);

        foreach (var arc in vChild.InArcs) {
          arc.Target = vClone;
          vClone.AddReverseArc(arc);
        }

        vClone.InArcs.Last().Data = vChild.InArcs.Last().Data; // fragment of child becomes fragment of clone
        GenealogyGraph.AddArc(vClone, vChild);

        // the following step in necessary because some mutation operators change values while the reference stays the same
        // so we clone in order to prevent mutations to be "shadowed" in the parent
        var parent = vClone.Parents.First();
        parent.Content = (T)parent.Content.Clone();
        GenealogyGraph[parent.Content] = parent;

      } else if (vChild.Rank.IsAlmost(Generations.Value)) {
        var clone = (T)ChildParameter.ActualValue.Clone();
        GenealogyGraph.SetContent(vChild, clone);
        var xx = new GenealogyGraphNode<T> { Content = ChildParameter.ActualValue, Rank = Generations.Value + 1 };
        GenealogyGraph.AddVertex(xx);
        GenealogyGraph.AddArc(vChild, xx);
      }
      return base.Apply();
    }
  }
}
